module ErrorProne


  # Mixed in to objects you wish to validate.
  #
  # @example
  #
  #   class FakeModel < Struct.new(:name)
  #     include ErrorProne::Model
  #   end
  #
  #   model = FakeModel.new
  #   model.valid?            # true
  #   model.add_error(:name, :too_cool)
  #   model.valid?            # false
  #   model.errors_for(:name) # [:too_cool]

  module Model

    # @param  [Symbol] field The field to find errors for
    # @return [Array] the errors for the given field
    def errors_for(field)
      errors.for(field)
    end

    # @return [ErrorProne::Errors]
    def errors
      @errors ||= Errors.new
    end

    # @param [Symbol] field Identifies which field to add an error to
    # @param [Symbol] error Identifies the type of error that occured
    # @return [self] The model itself
    def add_error(field, error)
      errors.add(field, error)
      self
    end

    # @return [Boolean] false if errors had been added. True if no errors.
    def valid?
      errors.empty?
    end
  end

  # Mixed in to objects that can verify whether an {ErrorProne::Model} is valid
  # or not.
  #
  # @example
  #   class FakeRule
  #
  #     include ErrorProne::Rule
  #     validates_as :not_nil
  #
  #     def verify!
  #       !value.nil?
  #     end
  #
  #     def message
  #       :field_nil
  #     end
  #   end
  #
  module Rule
    module ClassMethods

      # @param [Symbol] rule_name Name to use when using validates on an {ErrorProne::Validator}
      def validates_as(rule_name)
        RuleSet.available_rules[rule_name] = self
      end
    end


    attr_reader :object, :field, :value
    # @param [Object] object Object to verify conforms to a rule
    # @param [Symbol] field Field on the object that must conform to the rule
    # @param [Hash] options often specific to the the Rule
    def initialize(object, field, options = {})
      @object = object
      @field = field
      @options = options
    end

    # Adds error to the object if verify! fails
    # @return [Boolean]
    def validate!
      @value = object.send(field)
      return true if verify!
      object.add_error(field, message)
      false
    end

    # Message to add when validate! finds an error. Must override in
    # implementing classes.
    # @return [Symbol]
    def message
      raise "You must set a message for rules to work"
    end

    # Verifies if the rule is followed. Must override in implementing classes.
    # @return [Boolean]
    def verify!
      raise "You must provide a verify! method for rules to work. verify! should return a boolean"
    end

    def self.included(klazz)
      klazz.extend(ClassMethods)
    end
  end

  # Mixed in to objects you wish to use to validate other objects
  #
  # @example
  #   class ObjectToValidate < Struct.new(:name)
  #     include ErrorProne::Model
  #   end
  #
  #   class ObjectValidator
  #     include ErrorProne::Validator
  #     validates :name, :present
  #   end
  #
  #   object = ObjectToValidate.new(nil)
  #   ObjectValidator.new.validate!(object) # false
  #   object.valid? # false
  #   object.errors_for(:name) # [:missing]

  module Validator

    module ClassMethods
      # Contains rules the validator will run when .validate! is called
      #
      # @return (ErrorProne::RuleSet)
      def rules
        @rules ||= RuleSet.new
      end


      # Adds a rule the validator runs when .validate! is called.
      #
      # @param (Symbol) field Method to call on the object to validate
      # @param (Symbol) rule Name of rule to apply, defined by {ErrorProne::Rule}.validates_as
      # @param (Hash) options to pass to the {ErrorProne::Rule} when instantiating
      def validates(field, rule, options={})
        rules.add(field, rule, options)
      end

      # Validates all rules against the given object
      #
      # @param (Object) object Object to validate each rule against
      # @return (Boolean) True if all rules are valid, false otherwise.
      def validate!(object)
        rules.all_valid_for? object
      end
    end

    # Alias for {ErrorProne::Validator}.validate
    def validate!(object)
      self.class.validate!(object)
    end

    def self.included(base)
      base.extend(ClassMethods)
    end

  end

  class RuleSet < Hash
    def self.available_rules
      @available_rules ||= {}
    end

    def available_rules
      self.class.available_rules
    end

    def for(field)
      self[field] ||= []
    end

    def add(field, rule_name, options = {})
      raise "rule #{rule_name} is not available" unless available_rules.has_key?(rule_name)
      self.for(field).push(rule: available_rules[rule_name], options: options)
      self
    end


    def all_valid_for?(object)
      all? do |field, rules|
        rules.all? do |rule_config|
          rule = rule_config[:rule].new(object, field, rule_config[:options])
          rule.validate!
        end
      end
    end
  end

  class PresenceRule
    include Rule
    validates_as :present

    def verify!
      [:nil?, :empty?].select { |method| value.respond_to?(method) }.none? { |method|
        value.send(method)
      }
    end

    def message
      :missing
    end
  end

  class Errors < Hash
    def add(field, error)
      self.for(field).push(error)
      self
    end

    def for(field)
      self[field] ||= []
    end
  end
end
